; $Id: nat.scm,v 1.9 2006/05/03 14:43:39 thiel Exp $
(display "loading nat.scm ...") (newline)

(add-alg "nat" '("Zero" "nat") '("Succ" "nat=>nat"))

; Notice that nat.scm and numbers.scm cannot be loaded together.
; Reason: make-numeric-term (used by the parser) only gets a number.

(define (make-numeric-term n)
  (if (= n 0)
      (pt "Zero")
      (make-term-in-app-form
       (pt "Succ")
       (make-numeric-term (- n 1)))))

(define (is-numeric-term? term)
  (or
   (and (term-in-const-form? term)
	(string=? "Zero" (const-to-name (term-in-const-form-to-const term))))
   (and (term-in-app-form? term)
	(let ((op (term-in-app-form-to-op term)))
	  (and (term-in-const-form? op)
	       (string=? "Succ" (const-to-name
				 (term-in-const-form-to-const op)))
	       (is-numeric-term? (term-in-app-form-to-arg term)))))))

(define (numeric-term-to-number term)
  (if (equal? term (pt "Zero"))
      0
      (+ 1 (numeric-term-to-number (term-in-app-form-to-arg term)))))

(add-program-constant "NatPlus" (py "nat=>nat=>nat") t-deg-one)

(add-token
 "+" 'add-op
 (lambda (x y)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "NatPlus")) x y)))

(add-display
 (py "nat")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-final-op x))
	     (args (term-in-app-form-to-args x)))
	 (if (and (term-in-const-form? op)
		  (string=? "NatPlus"
			    (const-to-name (term-in-const-form-to-const op)))
		  (= 2 (length args)))
	     (list 'add-op "+"
		   (term-to-token-tree (car args))
		   (term-to-token-tree (cadr args)))
	     #f))
       #f)))

(add-computation-rule (pt "nat+0") (pt "nat"))
(add-computation-rule (pt "nat1+Succ nat2") (pt "Succ(nat1+nat2)"))

(quote (begin
(set-goal (pf "all nat 0+nat=nat"))
(ind)
(use "Truth-Axiom")
(assume "nat" "IH")
(use "IH")

(set-goal (pf "all nat1,nat2 Succ nat1+nat2=Succ(nat1+nat2)"))
(assume "nat1")
(ind)
(use "Truth-Axiom")
(assume "nat2" "IH")
(use "IH")

(set-goal (pf "all nat1,nat2,nat3 nat1+(nat2+nat3)=nat1+nat2+nat3"))
(assume "nat1" "nat2")
(ind)
(use "Truth-Axiom")
(assume "nat3" "IH")
(use "IH")
)) ;matches quote

(add-rewrite-rule (pt "0+nat") (pt "nat"))
(add-rewrite-rule (pt "Succ nat1+nat2") (pt "Succ(nat1+nat2)"))
(add-rewrite-rule (pt "nat1+(nat2+nat3)") (pt "nat1+nat2+nat3"))

(quote (begin
; "NatPlusComm"
(set-goal (pf "all nat1,nat2.nat1+nat2=nat2+nat1"))
(assume "nat1")
(ind)
(use "Truth-Axiom")
(assume "nat2" "IH")
(use "IH")
(save "NatPlusComm")
)) ;matches quote


(add-program-constant "NatTimes" (py "nat=>nat=>nat") t-deg-one)

(add-token
 "*" 'mul-op
 (lambda (x y)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "NatTimes")) x y)))

(add-display
 (py "nat")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-final-op x))
	     (args (term-in-app-form-to-args x)))
	 (if (and (term-in-const-form? op)
		  (string=? "NatTimes"
			    (const-to-name (term-in-const-form-to-const op)))
		  (= 2 (length args)))
	     (list 'mul-op "*"
		   (term-to-token-tree (car args))
		   (term-to-token-tree (cadr args)))
	     #f))
       #f)))

(add-computation-rule (pt "nat*0") (pt "0"))
(add-computation-rule (pt "nat1*Succ nat2") (pt "(nat1*nat2)+nat1"))

(quote (begin
; "NatTimesZero"
(set-goal (pf "all nat 0*nat=0"))
(ind)
(use "Truth-Axiom")
(assume "nat" "IH")
(use "IH")
(save "NatTimesZero")

; "NatCompat"
(set-goal
 (pf "all nat1,nat2.nat1=nat2 -> (Pvar nat)^nat1 -> (Pvar nat)^nat2"))
(ind)
(cases)
(assume "Trivil" "H1")
(use "H1")
(assume "nat" "Absurd" "H1")
(use "Efq")
(use "Absurd")
(assume "nat1" "IH")
(cases)
(assume "Absurd" "H1")
(use "Efq")
(use "Absurd")
(assume "nat2" "nat1=nat2")
(use-with "IH" (pt "nat2") "nat1=nat2" (pt "[nat]nat=>boole^(Succ nat)"))


(set-goal
 (pf "all nat1,nat2.nat1=nat2 -> 
                    all nat=>boole^.nat=>boole^nat1 -> nat=>boole^nat2"))
(ind)
(cases)
(assume "0=0" "nat=>boole^" "H1")
(use "H1")
(assume "nat" "Absurd" "nat=>boole^" "H1")
(use "Efq")
(use "Absurd")
(assume "nat1" "IH")
(cases)
(assume "Absurd" "nat=>boole^" "H1")
(use "Efq")
(use "Absurd")
(assume "nat2" "nat1=nat2" "nat=>boole^")
(use-with "IH" (pt "nat2") "nat1=nat2" (pt "[nat]nat=>boole^(Succ nat)"))
(save "Nat-Compat")

; "Nat-=-Compat"
(set-goal
 (pf "all nat1,nat2.nat1=nat2 -> 
                    all nat=>nat.nat=>nat nat1=nat=>nat nat2"))
(ind)
(cases)
(assume "Trivial" "nat=>nat")
(use "Truth-Axiom")
(assume "nat2" "Absurd" "nat=>nat")
(use "Efq")
(use "Absurd")
(assume "nat1" "IH")
(cases)
(assume "Absurd" "nat=>nat")
(use "Efq")
(use "Absurd")
(assume "nat2" "nat1=nat2" "nat=>nat")
(use-with "IH" (pt "nat2") "nat1=nat2" (pt "[nat]nat=>nat(Succ nat)"))
(save "Nat-=-Compat")

; "Nat-=-Comm"
(set-goal (pf "all nat1,nat2.nat1=nat2 -> nat2=nat1"))
(ind)
(cases)
(assume "H")
(use "H")
(assume "nat2" "Absurd")
(use "Absurd")
(assume "nat1" "IH")
(cases)
(assume "H")
(use "H")
(use "IH")
(save "Nat-=-Comm")

; "Nat-=-Trans"
(set-goal (pf "all nat1,nat2,nat3.nat1=nat2 -> nat2=nat3 -> nat1=nat3"))
(ind)
(cases)
(auto)
(strip)
(prop)
(assume "nat1" "IH")
(cases)
(strip)
(prop)
(assume "nat2")
(cases)
(prop)
(use "IH")
(save "Nat-=-Trans")

; "NatTimesSucc"
(set-goal (pf "all nat1,nat2 Succ nat1*nat2=(nat1*nat2)+nat2"))
(assume "nat1")
(ind)
(use "Truth-Axiom")
(assume "nat2" "IH")
(ng)
(use "Nat-=-Trans" (pt "nat1*nat2+nat2+nat1"))
(use-with "Nat-=-Compat" (pt "Succ nat1*nat2") (pt "nat1*nat2+nat2")
	  "IH" (pt "[nat]nat+nat1"))
(use-with "Nat-=-Compat" (pt "nat2+nat1") (pt "nat1+nat2") "?"
	  (pt "[nat]nat1*nat2+nat"))
(use "NatPlusComm")
(save "NatTimesSucc")

; "NatTimesComm"
(set-goal (pf "all nat1,nat2 nat1*nat2=nat2*nat1"))
(assume "nat1")
(ind)
(use "Nat-=-Comm")
(use "NatTimesZero")
(assume "nat2" "IH")
(ng)
(use "NatTimesSucc")
(use "Nat-=-Trans" (pt "nat2*nat1+nat1"))
(use-with "Nat-=-Compat" (pt "nat1*nat2") (pt "nat2*nat1") "IH"
	  (pt "[nat]nat+nat1"))
(use "Nat-=-Comm")
(use "NatTimesSucc")
(save "NatTimesComm")

; etc
)) ;matches quote above

(add-rewrite-rule (pt "0*nat") (pt "0"))
(add-rewrite-rule (pt "Succ nat1*nat2") (pt "(nat1*nat2)+nat2"))
(add-rewrite-rule (pt "nat1*(nat2*nat3)") (pt "nat1*nat2*nat3"))

(add-program-constant "NatLt" (py "nat=>nat=>boole") t-deg-one)

(add-token
 "<" 'rel-op
 (lambda (x y)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "NatLt")) x y)))

(add-display
 (py "boole")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-final-op x))
	     (args (term-in-app-form-to-args x)))
	 (if (and (term-in-const-form? op)
		  (string=? "NatLt"
			    (const-to-name (term-in-const-form-to-const op)))
		  (= 2 (length args)))
	     (list 'rel-op "<"
		   (term-to-token-tree (car args))
		   (term-to-token-tree (cadr args)))
	     #f))
       #f)))

(add-computation-rule (pt "nat<0") (pt "False"))
(add-computation-rule (pt "0<Succ nat") (pt "True"))
(add-computation-rule (pt "Succ nat1<Succ nat2") (pt "nat1<nat2"))

(quote (begin
(set-goal (pf "all nat nat<Succ nat"))
(ind)
(use "Truth-Axiom")
(assume "nat" "IH")
(use "IH")

(set-goal (pf "all nat.nat<nat -> F"))
(ind)
(assume "H1")
(use "H1")
(assume "nat" "IH")
(use "IH")

(set-goal (pf "all nat.Succ nat<nat -> F"))
(ind)
(assume "H1")
(use "H1")
(assume "nat" "IH")
(use "IH")
)) ;matches quote

(add-rewrite-rule (pt "nat<Succ nat") (pt "True"))
; The following rule is temporarily removed, for compatibility
; (add-rewrite-rule (pt "nat<nat") (pt "False")) 
(add-rewrite-rule (pt "Succ nat<nat") (pt "False"))

; "NatLtTrans"
(set-goal (pf "all nat1,nat2,nat3.nat1<nat2 -> nat2<nat3 -> nat1<nat3"))
(ind)
(ind)
(strip)
(use 2)
(assume "nat2" "IH2")
(cases)
(prop)
(strip)
(prop)
(assume "nat1" "IH1")
(ind)
(strip)
(prop)
(assume "nat2" "IH2")
(cases)
(prop)
(assume "nat3")
(use "IH1")
(save "NatLtTrans")

; "NatLtSuccCases"
(set-goal (pf "all nat2,nat1.nat1<Succ nat2 -> (nat1<nat2 -> Pvar^) -> 
                             (nat1=nat2 -> Pvar^) -> Pvar^"))
(ind)
(cases)
(assume "H1" "H2" "H3")
(use "H3")
(use "Truth-Axiom")
(assume "nat1" "Absurd" "H2" "H3")
(use "Efq")
(use "Absurd")
(assume "nat1" "IH")
(cases)
(assume "H1" "H2" "H3")
(use "H2")
(use "Truth-Axiom")
(use "IH")
(save "NatLtSuccCases")


(add-program-constant "NatLe" (py "nat=>nat=>boole") t-deg-one)

(add-token
 "<=" 'rel-op
 (lambda (x y)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "NatLe")) x y)))

(add-display
 (py "boole")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-final-op x))
	     (args (term-in-app-form-to-args x)))
	 (if (and (term-in-const-form? op)
		  (string=? "NatLe"
			    (const-to-name (term-in-const-form-to-const op)))
		  (= 2 (length args)))
	     (list 'rel-op "<="
		   (term-to-token-tree (car args))
		   (term-to-token-tree (cadr args)))
	     #f))
       #f)))

(add-computation-rule (pt "0<=nat") (pt "True"))
(add-computation-rule (pt "Succ nat<=0") (pt "False"))
(add-computation-rule (pt "Succ nat1<=Succ nat2") (pt "nat1<=nat2"))

(quote (begin
(set-goal (pf "all nat nat<=nat"))
(ind)
(use "Truth-Axiom")
(assume "nat" "IH")
(use "IH")

(set-goal (pf "all nat1,nat2 nat1<=nat1+nat2"))
(ind)
(assume "nat2")
(use "Truth-Axiom")
(assume "nat" "IH")
(use "IH")

(set-goal (pf "all nat.Succ nat<=nat -> F"))
(ind)
(prop)
(assume "nat" "IH")
(use "IH")
)) ;matches quote


(add-rewrite-rule (pt "nat<=nat") (pt "True"))
(add-rewrite-rule (pt "nat1<=nat1+nat2") (pt "True"))
(add-rewrite-rule (pt "Succ nat<=nat") (pt "False"))

; "NatLeTrans"
(set-goal (pf "all nat1,nat2,nat3.nat1<=nat2 -> nat2<=nat3 -> nat1<=nat3"))
(ind)
(ind)
(strip)
(use 2)
(assume "nat2" "IH2")
(cases)
(strip)
(use "Truth-Axiom")
(strip)
(use "Truth-Axiom")
(assume "nat1" "IH1")
(ind)
(strip)
(use "Efq")
(use 2)
(assume "nat2" "IH2")
(cases)
(strip)
(use 4)
(assume "nat3")
(use "IH1")
(save "NatLeTrans")

; "NatLtLeTrans"
(set-goal (pf "all nat1,nat2,nat3.nat1<nat2 -> nat2<=nat3 -> nat1<nat3"))
(ind)
(ind)
(strip)
(use "Efq")
(use 1)
(assume "nat2" "IH2")
(cases)
(strip)
(use 3)
(strip)
(use "Truth-Axiom")
(assume "nat1" "IH1")
(ind)
(strip)
(use "Efq")
(use 2)
(assume "nat2" "IH2")
(cases)
(strip)
(use 4)
(assume "nat3")
(use "IH1")
(save "NatLtLeTrans")

; "NatLeLtTrans"
(set-goal (pf "all nat1,nat2,nat3.nat1<=nat2 -> nat2<nat3 -> nat1<nat3"))
(ind)
(ind)
(strip)
(use 2)
(assume "nat2" "IH2")
(cases)
(strip)
(use 3)
(strip)
(use "Truth-Axiom")
(assume "nat1" "IH1")
(ind)
(strip)
(use "Efq")
(use 2)
(assume "nat2" "IH2")
(cases)
(strip)
(use 4)
(assume "nat3")
(use "IH1")
(save "NatLeLtTrans")

; "NatLtSuccToLe"
(set-goal (pf "all nat1,nat2.nat1<Succ nat2 -> nat1<=nat2"))
(ind)
(cases)
(strip)
(use "Truth-Axiom")
(strip)
(use "Truth-Axiom")
(assume "nat1" "IH1")
(cases)
(assume "Absurd")
(use "Absurd")
(assume "nat2")
(use "IH1")
(save "NatLtSuccToLe")

; "NatLtLtSuccTrans"
(set-goal (pf "all nat1,nat2,nat3.nat1<nat2 -> nat2<Succ nat3 -> nat1<nat3"))
(assume "nat1" "nat2" "nat3" "nat1<nat2" "nat2<Succ nat3")
(use "NatLtLeTrans" (pt "nat2"))
(use "nat1<nat2")
(use "NatLtSuccToLe")
(use "nat2<Succ nat3")
(save "NatLtLtSuccTrans")

; "NatLeToLtSucc"
(set-goal (pf "all nat1,nat2.nat1<=nat2 -> nat1<Succ nat2"))
(ind)
(cases)
(strip)
(use "Truth-Axiom")
(strip)
(use "Truth-Axiom")
(assume "nat1" "IH1")
(cases)
(assume "Absurd")
(use "Absurd")
(assume "nat2")
(use "IH1")
(save "NatLeToLtSucc")

; "NatLeCases"
(set-goal (pf "all nat2,nat1.nat1<=nat2 -> (nat1<nat2 -> Pvar^) -> 
                             (nat1=nat2 -> Pvar^) -> Pvar^"))
(ind)
(cases)
(assume "H1" "H2" "H3")
(use "H3")
(use "Truth-Axiom")
(assume "nat1" "Absurd" "H2" "H3")
(use "Efq")
(use "Absurd")
(assume "nat1" "IH")
(cases)
(assume "H1" "H2" "H3")
(use "H2")
(use "Truth-Axiom")
(use "IH")
(save "NatLeCases")


(add-program-constant "Pred" (py "nat=>nat") t-deg-one)

(add-computation-rule (pt "Pred 0") (pt "0"))
(add-computation-rule (pt "Pred(Succ nat)") (pt "nat"))

(add-program-constant "NatMinus" (py "nat=>nat=>nat") t-deg-one)

(add-token
 "-" 'add-op
 (lambda (x y)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "NatMinus")) x y)))

(add-display
 (py "nat")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-final-op x))
	     (args (term-in-app-form-to-args x)))
	 (if (and (term-in-const-form? op)
		  (string=? "NatMinus"
			    (const-to-name (term-in-const-form-to-const op)))
		  (= 2 (length args)))
	     (list 'add-op "-"
		   (term-to-token-tree (car args))
		   (term-to-token-tree (cadr args)))
	     #f))
       #f)))

(add-computation-rule (pt "nat-0") (pt "nat"))
(add-computation-rule (pt "nat1-Succ nat2") (pt "Pred(nat1-nat2)"))

(quote (begin

; "NatPredMinusSucc"
(set-goal (pf "all nat1,nat2 Pred(Succ nat1-nat2)=nat1-nat2"))
(assume "nat1")
(ind)
(use "Truth-Axiom")
(assume "nat2" "IH")
(ng)
(use-with "Nat-=-Compat"
	  (pt "Pred(Succ nat1-nat2)")
	  (pt "nat1-nat2")
	  "IH"
	  (pt "[nat]Pred nat"))
(save "NatPredMinusSucc")

(set-goal (pf "all nat nat-nat=0"))
(ind)
(use "Truth-Axiom")
(assume "nat" "IH")
(ng)
(use "Nat-=-Trans" (pt "nat-nat"))
(use "NatPredMinusSucc")
(use "IH")

(set-goal (pf "all nat.Succ nat<=nat -> F"))
(ind)
(prop)
(assume "nat" "IH")
(use "IH")

)) ;matches quote

(add-rewrite-rule (pt "Pred(Succ nat1-nat2)") (pt "nat1-nat2"))
(add-rewrite-rule (pt "nat-nat") (pt "0"))
(add-rewrite-rule (pt "Succ nat-nat") (pt "1"))


(add-program-constant "NatMax" (py "nat=>nat=>nat") t-deg-one)

(add-token
 "max" 'mul-op
 (lambda (x y)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "NatMax")) x y)))

(add-display
 (py "nat")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-final-op x))
	     (args (term-in-app-form-to-args x)))
	 (if (and (term-in-const-form? op)
		  (string=? "NatMax"
			    (const-to-name (term-in-const-form-to-const op)))
		  (= 2 (length args)))
	     (list 'mul-op "max"
		   (term-to-token-tree (car args))
		   (term-to-token-tree (cadr args)))
	     #f))
       #f)))

(add-computation-rule (pt "nat max 0") (pt "nat"))
(add-computation-rule (pt "0 max Succ nat") (pt "Succ nat"))
(add-computation-rule (pt "Succ nat1 max Succ nat2")
		      (pt "Succ(nat1 max nat2)"))

(add-rewrite-rule (pt "0 max nat") (pt "nat"))
(add-rewrite-rule (pt "nat max nat") (pt "nat"))
(add-rewrite-rule (pt "nat1 max (nat2 max nat3)")
		  (pt "nat1 max nat2 max nat3"))

; "NatMaxUB1" 
(set-goal (pf "all nat1,nat2 nat1<=nat1 max nat2"))
(ind)
(assume "nat2")
(use "Truth-Axiom")
(assume "nat1" "IH1")
(cases)
(use "Truth-Axiom")
(use "IH1")
(save "NatMaxUB1")

; "NatMaxUB2" 
(set-goal (pf "all nat1,nat2 nat2<=nat1 max nat2"))
(ind)
(assume "nat2")
(use "Truth-Axiom")
(assume "nat1" "IH1")
(cases)
(use "Truth-Axiom")
(use "IH1")
(save "NatMaxUB2")


(add-program-constant "NatMin" (py "nat=>nat=>nat") t-deg-one)

(add-token
 "min" 'mul-op
 (lambda (x y)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "NatMin")) x y)))

(add-display
 (py "nat")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-final-op x))
	     (args (term-in-app-form-to-args x)))
	 (if (and (term-in-const-form? op)
		  (string=? "NatMin"
			    (const-to-name (term-in-const-form-to-const op)))
		  (= 2 (length args)))
	     (list 'mul-op "min"
		   (term-to-token-tree (car args))
		   (term-to-token-tree (cadr args)))
	     #f))
       #f)))

(add-computation-rule (pt "nat min 0") (pt "0"))
(add-computation-rule (pt "0 min Succ nat") (pt "0"))
(add-computation-rule (pt "Succ nat1 min Succ nat2")
		      (pt "Succ(nat1 min nat2)"))

(add-rewrite-rule (pt "0 min nat") (pt "0"))
(add-rewrite-rule (pt "nat min nat") (pt "nat"))
(add-rewrite-rule (pt "nat1 min (nat2 min nat3)")
		  (pt "nat1 min nat2 min nat3"))

; "NatMinLB1" 
(set-goal (pf "all nat1,nat2 nat1 min nat2<=nat1"))
(ind)
(assume "nat2")
(use "Truth-Axiom")
(assume "nat1" "IH1")
(cases)
(use "Truth-Axiom")
(use "IH1")
(save "NatMinLB1")

; "NatMinLB2" 
(set-goal (pf "all nat1,nat2 nat1 min nat2<=nat2"))
(ind)
(assume "nat2")
(use "Truth-Axiom")
(assume "nat1" "IH1")
(cases)
(use "Truth-Axiom")
(use "IH1")
(save "NatMinLB2")


(add-program-constant "AllBNat" (py "nat=>(nat=>boole)=>boole") t-deg-one)

(add-computation-rule (pt "AllBNat 0 nat=>boole") (pt "True"))
(add-computation-rule (pt "AllBNat(Succ nat)nat=>boole")
		      (pt "AllBNat nat nat=>boole andb (nat=>boole)nat"))

; "AllBNatIntro"
(set-goal
 (pf "all nat=>boole^,nat2.(all nat1.nat1<nat2 -> nat=>boole^nat1) ->
                                     AllBNat nat2([nat1]nat=>boole^nat1)"))
(assume "nat=>boole^")
(ind)
(ng)
(strip)
(use "Truth-Axiom")
(assume "nat2" "IH")
(ng)
(assume "H")
(split)
(use "IH")
(assume "nat1" "nat1<nat2")
(use "H")
(use "NatLtTrans" (pt "nat2"))
(use "nat1<nat2")
(use "Truth-Axiom")
(use "H")
(use "Truth-Axiom")
(save "AllBNatIntro")

; "AllBNatElim"
(set-goal
 (pf "all nat=>boole^,nat2.AllBNat nat2 nat=>boole^ -> 
                           all nat1.nat1<nat2 -> nat=>boole^ nat1"))
(assume "nat=>boole^")
(ind)
(ng)
(strip)
(use "Efq")
(use 2)
(assume "nat2" "IH")
(ng)
(assume "AllBHyp" "nat1" "nat1<Succ nat2")
(use "NatLtSuccCases" (pt "nat2") (pt "nat1"))
(use "nat1<Succ nat2")
(use "IH")
(use "AllBHyp")
(assume "nat1=nat2")
(simp "nat1=nat2")
(use "AllBHyp")
(save "AllBNatElim")

(av "n" "m" "k" (py "nat"))
