; $Id: numbers.scm,v 1.6 2006/01/02 09:12:08 schwicht Exp $

; Based on ~schwicht/wwwpublic/seminars/prosemss04/numbers.scm,v 1.23 
; 2004/11/02

; Changed to work with binary numbers, for efficiency reasons.

(display "loading numbers.scm ...") (newline)


; 1. Positive Numbers, Integers, Rationals and Reals
; ==================================================

; We want to overload operators like +,*,/ <=,<, and also
; automatically raise the type of (coerce) arguments.  E.g. + should
; take its arguments, compute the lub rho of their types, raise the
; type of both arguments to type rho, apply RhoPlus to the results.

; The code for positive numbers is from
; wwwpublic/lectures/logic/ws03/pos.scm

(add-alg "pos" '("One" "pos") '("SZero" "pos=>pos") '("SOne" "pos=>pos"))

; An integer is either positive or zero or negative.

(add-alg
 "int" '("IntPos" "pos=>int") '("IntZero" "int") '("IntNeg" "pos=>int"))

; make-numeric-term produces an int object for 0, and a pos object for
; a positive integer.

(define (make-numeric-term n) 
  (cond ((zero? n) (make-term-in-const-form (constr-name-to-constr "IntZero")))
	((= n 1) (pt "One"))
	((even? n) (make-term-in-app-form
		    (make-term-in-const-form (constr-name-to-constr "SZero"))
		    (make-numeric-term (/ n 2))))
	((odd? n) (make-term-in-app-form
		   (make-term-in-const-form (constr-name-to-constr "SOne"))
		   (make-numeric-term (/ (- n 1) 2))))
	(else (myerror "make-numeric-term" "natural number expected" n))))

(define (is-numeric-term? term)
  (or
   (and (term-in-const-form? term)
	(string=? "IntZero"
		  (const-to-name (term-in-const-form-to-const term))))
   (and (term-in-const-form? term)
	(string=? "One" (const-to-name (term-in-const-form-to-const term))))
   (and (term-in-app-form? term)
	(let ((op (term-in-app-form-to-op term))
	      (arg (term-in-app-form-to-arg term)))
	  (and (term-in-const-form? op)
	       (let ((name (const-to-name (term-in-const-form-to-const op))))
		 (or (and (string=? "SZero" name) (is-numeric-term? arg))
		     (and (string=? "SOne" name) (is-numeric-term? arg)))))))))

(define (numeric-term-to-number term)
  (cond
   ((and (term-in-const-form? term)
	 (string=? "IntZero" (const-to-name
			      (term-in-const-form-to-const term))))
    0)
   ((and (term-in-const-form? term)
	 (string=? "One" (const-to-name
			  (term-in-const-form-to-const term))))
    1)
   ((term-in-app-form? term)
    (let ((op (term-in-app-form-to-op term))
	  (arg (term-in-app-form-to-arg term)))
      (if
       (term-in-const-form? op)
       (let ((name (const-to-name (term-in-const-form-to-const op))))
	 (cond
	  ((string=? "SZero" name)
	   (* 2 (numeric-term-to-number arg)))
	  ((string=? "SOne" name)
	   (+ 1 (* 2 (numeric-term-to-number arg))))
	  (else (myerror "numeric-term-to-number" "unexpected term"
			 (term-to-string term)))))
       (myerror "numeric-term-to-number" "unexpected term"
		(term-to-string term)))))
   (else (myerror "numeric-term-to-number" "unexpected term"
		  (term-to-string term)))))

; (term-to-string (make-numeric-term 0))
; (term-to-string (make-numeric-term 270))
; (term-to-string (pt "270"))

; We postpone introduction of + etc, and first introduce the types rat
; of rationals and real of reals.  A rational is a pair of an integer
; and a positive number.

; Comparison with the treatment of constructive analysis in Nijmegen
; C-CoRN.  They want to be as general as possible, and e.g. derive
; properties of the rationals by instantiation of general field facts.
; However, when one wants to use proofs for program extraction and
; development, it seems better to have a more direct approach, which
; allows to take care of the underlying data structures.

(add-var-name "n" "m" "k" "l" "p" "q" (py "pos"))
(add-var-name "i" "j" (py "int"))

(add-token
 "IntN"
 'prefix-op
 (lambda (x) (mk-term-in-app-form (pt "IntNeg") x)))

(add-token
 "IntP"
 'prefix-op
 (lambda (x) (mk-term-in-app-form (pt "IntPos") x)))

(add-display
 (py "int")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (= 1 (length args)))
	 (let ((name (const-to-name (term-in-const-form-to-const op))))
	   (cond
	    ((string=? name "IntNeg")
	     (list 'prefix-op "IntN" (term-to-token-tree (car args))))
	    ((string=? name "IntPos")
	     (term-to-token-tree (car args)))
	    (else #f)))
	 #f))))

; (term-to-string (pt "IntP One"))
; (term-to-string (pt "IntN One"))

; For systematic reasons we want the following maps
; pos-to-pos-term                  not needed, use make-numeric-term
; pos-numeric-term-to-number       not needed, use numeric-term-to-number
; is-pos-numeric-term?

; pos-term-to-int-term
; int-term-to-pos-term

; int-to-int-term
; int-numeric-term-to-number
; is-int-numeric-term?

; int-term-to-rat-term
; rat-term-to-int-term

; rat-to-rat-term
; rat-numeric-term-to-number
; is-rat-numeric-term?

(define (is-pos-numeric-term? term)
  (and (is-numeric-term? term)
       (positive? (numeric-term-to-number term))))

(define (pos-term-to-int-term term)
  (mk-term-in-app-form
   (make-term-in-const-form (constr-name-to-constr "IntPos"))
   term))

(define (int-term-to-pos-term term)
  (let ((op (term-in-app-form-to-final-op term))
	(args (term-in-app-form-to-args term)))
    (if (and (term-in-const-form? op)
	     (string=? "IntPos"
		       (const-to-name (term-in-const-form-to-const op)))
	     (= 1 (length args)))
	(car args)
	term)))

(define (int-to-int-term n)
  (cond
   ((positive? n) (pos-term-to-int-term (make-numeric-term n)))
   ((zero? n) (make-numeric-term 0))
   ((negative? n)
    (mk-term-in-app-form
     (make-term-in-const-form (constr-name-to-constr "IntNeg"))
     (make-numeric-term (- n))))
   (else (myerror "int-to-int-term" "integer expected" n))))

; (term-to-string (int-to-int-term 2))
; (term-to-string (int-to-int-term 0))
; (term-to-string (int-to-int-term -2))

(define (int-numeric-term-to-number term)
  (cond
   ((and (term-in-const-form? term)
	 (string=? "IntZero" (const-to-name
			      (term-in-const-form-to-const term))))
    0)
   ((term-in-app-form? term)
    (let ((op (term-in-app-form-to-op term))
	  (arg (term-in-app-form-to-arg term)))
      (if (term-in-const-form? op)
	  (let ((name (const-to-name (term-in-const-form-to-const op))))
	    (cond
	     ((string=? name "IntPos") (numeric-term-to-number arg))
	     ((string=? name "IntNeg") (- (numeric-term-to-number arg)))
	     (else (myerror "int-numeric-term-to-number"
			    "int numeric term expected"
			    (term-to-string term)))))
	  (myerror "int-numeric-term-to-number" "int numeric term expected"
		   (term-to-string term)))))
   (else (myerror "int-numeric-term-to-number" "int numeric term expected"
		  (term-to-string term)))))
  
; (int-numeric-term-to-number (int-to-int-term 2))
; (int-numeric-term-to-number (int-to-int-term 0))
; (int-numeric-term-to-number (int-to-int-term -22))
	
(define (is-int-numeric-term? term)
  (or
   (and
    (term-in-const-form? term)
    (string=? "IntZero" (const-to-name
			 (term-in-const-form-to-const term))))
   (and
    (term-in-app-form? term)
    (let ((op (term-in-app-form-to-op term))
	  (arg (term-in-app-form-to-arg term)))
      (and (term-in-const-form? op)
	   (let ((name (const-to-name (term-in-const-form-to-const op))))
	     (or
	      (and (string=? name "IntPos") (is-pos-numeric-term? arg))
	      (and (string=? name "IntNeg") (is-pos-numeric-term? arg)))))))))
		 
; (is-int-numeric-term? (int-to-int-term 2))
; (is-int-numeric-term? (int-to-int-term 0))
; (is-int-numeric-term? (int-to-int-term -22))

; 04-09-03 To use external code in a pconst, we provide tests for
; numeral values and conversion operations from numerals values to
; numbers, both for pos and int

(define (pos-numeral-value? value)
  (and (nbe-constr-value? value)
       (let* ((name (nbe-constr-value-to-name value))
	      (args (nbe-constr-value-to-args value))
	      (vals (map nbe-object-to-value args)))
	 (or (string=? "One" name)
	     (and (or (string=? "SZero" name) (string=? "SOne" name))
		  (pair? vals)
		  (pos-numeral-value? (car vals)))))))

(define (pos-numeral-value-to-number value)
  (let* ((name (nbe-constr-value-to-name value))
	 (args (nbe-constr-value-to-args value))
	 (vals (map nbe-object-to-value args)))
    (if (string=? "One" name)
	1
	(let ((prev (pos-numeral-value-to-number (car vals))))
	  (if (string=? "SZero" name) (* 2 prev) (+ (* 2 prev) 1))))))

; Instead of the converse use pt.

(define (int-numeral-value? value)
  (and (nbe-constr-value? value)
       (let* ((name (nbe-constr-value-to-name value))
	      (args (nbe-constr-value-to-args value))
	      (vals (map nbe-object-to-value args)))
	 (or (and (string=? "IntPos" name)
		  (pos-numeral-value? (car vals)))
	     (string=? "IntZero" name)
	     (and (string=? "IntNeg" name)
		  (pos-numeral-value? (car vals)))))))

(define (int-numeral-value-to-number value)
  (let* ((name (nbe-constr-value-to-name value))
	 (args (nbe-constr-value-to-args value))
	 (vals (map nbe-object-to-value args)))
    (cond
     ((string=? "IntNeg" name) (- (pos-numeral-value-to-number (car vals))))
     ((string=? "IntZero" name) 0)
     ((string=? "IntPos" name) (pos-numeral-value-to-number (car vals)))
     (else (myerror "int-numeral-value-to-number" "unexpected arg" value)))))


; We also introduce the rationals.  A rational is a pair i#n of an
; integer i and a positive natural number n, interpreted as i/n.

(add-alg "rat" '("RatConstr" "int=>pos=>rat"))
(add-var-name "a" "b" "c" "d" (py "rat"))

(add-token
 "#"
 'pair-op
 (lambda (x y)
   (mk-term-in-app-form 
    (make-term-in-const-form (constr-name-to-constr "RatConstr"))
    x y)))

(add-display
 (py "rat")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RatConstr"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (if (and (term-in-const-form? (cadr args))
		  (string=? "One" (const-to-name
				   (term-in-const-form-to-const (cadr args)))))
	     (term-to-token-tree (car args))
	     (list 'pair-op "#"
		   (term-to-token-tree (car args))
		   (term-to-token-tree (cadr args))))
	 #f))))

; int-term-to-rat-term
; rat-term-to-int-term

; rat-to-rat-term
; rat-numeric-term-to-number
; is-rat-numeric-term?

(define (int-term-to-rat-term term)
  (mk-term-in-app-form
   (make-term-in-const-form (constr-name-to-constr "RatConstr"))
   term (make-numeric-term 1)))

; (term-to-string (int-term-to-rat-term (int-to-int-term -90)))
  
(define (rat-term-to-int-term term)
  (let ((op (term-in-app-form-to-final-op term))
	(args (term-in-app-form-to-args term)))
    (if (and (term-in-const-form? op)
	     (string=? "RatConstr" (const-to-name
				    (term-in-const-form-to-const op)))
	     (= 2 (length args))
	     (is-numeric-term? (cadr args))
	     (= 1 (numeric-term-to-number (cadr args))))
	(car args)
	term)))

(define (rat-to-rat-term n)
  (mk-term-in-app-form
   (make-term-in-const-form (constr-name-to-constr "RatConstr"))
   (int-to-int-term (numerator n))
   (make-numeric-term (denominator n))))

; (term-to-string (rat-to-rat-term -2/3))
; (term-to-string (rat-to-rat-term 2/3))
; (term-to-string (rat-to-rat-term 21/3))

(define (rat-numeric-term-to-number term)
  (let* ((args (term-in-app-form-to-args term))
	 (k (numeric-term-to-number (cadr args)))
	 (i (int-numeric-term-to-number (car args))))
    (/ i k)))

; (rat-numeric-term-to-number (rat-to-rat-term -2/3))
; (rat-numeric-term-to-number (pt "IntN 2#3"))
; (rat-numeric-term-to-number (rat-to-rat-term 7/3))

(define (is-rat-numeric-term? term)
  (and (term-in-app-form? term)
       (let ((op (term-in-app-form-to-final-op term))
	     (args (term-in-app-form-to-args term)))
	 (and (term-in-const-form? op)
	      (string=? "RatConstr"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args))
	      (let ((arg1 (car args))
		    (arg2 (cadr args)))
		(and (is-int-numeric-term? arg1)
		     (is-pos-numeric-term? arg2)))))))

; (is-rat-numeric-term? (rat-to-rat-term -2/3))

(define (pos-term-to-rat-term term)
  (int-term-to-rat-term (pos-term-to-int-term term)))

; (term-to-string (pos-term-to-rat-term (pt "3")))


; We now introduce the reals.  A real is a pair of a Cauchy sequence
; of rationals and a modulus.  We view real as a data type (i.e., no
; properties), and within this data type inductively define the
; predicate Real x, meaning that x is a (proper) real.

(add-alg "real" (list "RealConstr" "(pos=>rat)=>(pos=>pos)=>real"))

(add-var-name "as" "bs" "cs" "ds" (py "pos=>rat"))
(add-var-name "M" "N" (py "pos=>pos"))
(add-var-name "x" "y" "z" (py "real"))

; To conveniently access the two fields of a real, we provide seq and
; mod as informative names to be used (postfix) in display strings.

(add-program-constant "RealSeq" (py "real=>pos=>rat") 1 'const 1)

(add-token
 "seq"
 'postfix-op
 (lambda (x)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "RealSeq"))
    x)))

(add-display
 (py "pos=>rat")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealSeq"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 1 (length args)))
	 (let ((arg (car args)))
	   (list
	    'postfix-op "seq"
	    (term-to-token-tree arg)))
	 #f))))
 
(add-computation-rule (pt "(RealConstr as M)seq") (pt "as"))

(add-program-constant "RealMod" (py "real=>pos=>pos") 1 'const 1)

(add-token
 "mod"
 'postfix-op
 (lambda (x)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "RealMod"))
    x)))

(add-display
 (py "pos=>pos")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealMod"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 1 (length args)))
	 (let ((arg (car args)))
	   (list
	    'postfix-op "mod"
	    (term-to-token-tree arg)))
	 #f))))
 
(add-computation-rule (pt "(RealConstr as M)mod") (pt "M"))

; (term-to-string (pt "x seq n"))
; (term-to-string (pt "x mod n"))

; We now provide some coercion functions.

(define (rat-term-to-real-term term)
  (let ((var (type-to-new-var (py "pos"))))
    (mk-term-in-app-form
     (make-term-in-const-form (constr-name-to-constr "RealConstr"))
     (make-term-in-abst-form var term) ;constant Cauchy sequence
     (make-term-in-abst-form var (make-numeric-term 1))))) ;one modulus

; (term-to-string (rat-term-to-real-term (rat-to-rat-term -2/3)))
; "RealConstr([n14]IntN 2#3)([n14]1)"
; (term-to-string (rat-term-to-real-term (pt "a")))

(define (int-term-to-real-term term)
  (rat-term-to-real-term (int-term-to-rat-term term)))

(define (pos-term-to-real-term term)
  (rat-term-to-real-term (int-term-to-rat-term (pos-term-to-int-term term))))

(define (intreal-term-to-real-term term)
  (let* ((seq-const (pconst-name-to-pconst "RealSeq"))
	 (mod-const (pconst-name-to-pconst "RealMod")))
    (mk-term-in-app-form
     (make-term-in-const-form (constr-name-to-constr "RealConstr"))
     (intseq-term-to-ratseq-term
      (make-term-in-app-form (make-term-in-const-form seq-const) term))
     (make-term-in-app-form (make-term-in-const-form mod-const) term))))

(define (intseq-term-to-ratseq-term term)
  (let* ((pos (py "pos"))
	 (var (type-to-new-var pos)))
    (make-term-in-abst-form
     var (int-term-to-rat-term
	  (make-term-in-app-form term (make-term-in-var-form var))))))

; (term-to-string (nt (intseq-term-to-ratseq-term (pt "[n]i"))))

(define (posseq-term-to-intseq-term term)
  (let* ((pos (py "pos"))
	 (var (type-to-new-var pos)))
    (make-term-in-abst-form
     var (pos-term-to-int-term
	  (make-term-in-app-form term (make-term-in-var-form var))))))

(define (posseq-term-to-ratseq-term term)
  (let* ((pos (py "pos"))
	 (var (type-to-new-var pos)))
    (make-term-in-abst-form
     var (pos-term-to-rat-term
	  (make-term-in-app-form term (make-term-in-var-form var))))))

; and also downwards:

(define (real-term-to-rat-term term)
  (let ((op (term-in-app-form-to-final-op term))
	(args (term-in-app-form-to-args term)))
    (if (and (term-in-const-form? op)
	     (string=? "RealConstr" (const-to-name
				     (term-in-const-form-to-const op)))
	     (= 2 (length args)))
	(let ((arg1 (car args))
	      (arg2 (cadr args)))
	  (if (and (term-in-abst-form? arg1))
	      (let ((var (term-in-abst-form-to-var arg1))
		    (kernel (term-in-abst-form-to-kernel arg1)))
		 (if (not (member var (term-to-free kernel)))
		     kernel
		     term))
	      term))
	term)))

(set! INCREASING-ALGEBRAS-TO-COERCION-ALIST
      (cons (list (list (make-alg "pos") (make-alg "int"))
		  pos-term-to-int-term)
	    INCREASING-ALGEBRAS-TO-COERCION-ALIST))

(set! INCREASING-ALGEBRAS-TO-COERCION-ALIST
      (cons (list (list (make-alg "pos") (make-alg "rat"))
		  pos-term-to-rat-term)
	    INCREASING-ALGEBRAS-TO-COERCION-ALIST))

(set! INCREASING-ALGEBRAS-TO-COERCION-ALIST
      (cons (list (list (make-alg "pos") (make-alg "real"))
		  pos-term-to-real-term)
	    INCREASING-ALGEBRAS-TO-COERCION-ALIST))

(set! INCREASING-ALGEBRAS-TO-COERCION-ALIST
      (cons (list (list (make-alg "int") (make-alg "rat"))
		  int-term-to-rat-term)
	    INCREASING-ALGEBRAS-TO-COERCION-ALIST))

(set! INCREASING-ALGEBRAS-TO-COERCION-ALIST
      (cons (list (list (make-alg "int") (make-alg "real"))
		  int-term-to-real-term)
	    INCREASING-ALGEBRAS-TO-COERCION-ALIST))

(set! INCREASING-ALGEBRAS-TO-COERCION-ALIST
      (cons (list (list (make-alg "rat") (make-alg "real"))
		  rat-term-to-real-term)
	    INCREASING-ALGEBRAS-TO-COERCION-ALIST))

; (term-to-string (real-term-to-rat-term (nt (pt "RealConstr([n]i)([n]7)"))))

; We will also need

(define (rat-to-real-term n)
  (rat-term-to-real-term (rat-to-rat-term n)))

; (term-to-string (rat-to-real-term 2))
; (term-to-string (rat-to-real-term -2/3))

(define (is-real-numeric-term? term)
  (let ((op (term-in-app-form-to-final-op term))
	(args (term-in-app-form-to-args term)))
    (and (term-in-const-form? op)
	 (string=? "RealConstr"
		   (const-to-name (term-in-const-form-to-const op)))
	 (= 2 (length args))
	 (let ((arg1 (car args)))
	   (and (term-in-abst-form? arg1)
		(is-rat-numeric-term?
		 (term-in-abst-form-to-kernel arg1)))))))

; (is-real-numeric-term? (rat-to-real-term 2))

(define (real-numeric-term-to-number term)
  (let ((args (term-in-app-form-to-args term)))
    (if (= 2 (length args))
	(rat-numeric-term-to-number
	 (if (term-in-abst-form? (car args))
	     (term-in-abst-form-to-kernel (car args))
	     (myerror "real-numeric-term-to-number"
		      "term in abst form expected"
		      (term-to-string (car args)))))
	(myerror "real-numeric-term-to-number" "real numeric term expected"
		 (term-to-string term)))))

; (real-numeric-term-to-number (rat-to-real-term 2))


; 2. Parsing and Display for Arithmetical Operations
; ==================================================

; We now come to some generalities concerning overloading and coercion.
; Requirements:
; - Internally every application is type correct
; - Displaying a term can lower its type, from real to rat to int to pos
; - Parsing (1) a term and (2) a new term obtained from it by lowering
;   the type of some of its subterms must always return the same result,
;   possibly up to lowering of its type, from rat to int to nat to pos.

; Possible problems with usage of terms as arguments of predicates:
; (P(3#1) is displayed as P 3, which is not type correct) or in lists:
; (3#1)::(1#2): is displayed as 3::(1#2): , which is not type correct.

; Solution: change make-term-in-app-form and make-predicate-formula.
; If the types do not fit, raise the types of the offending arguments
; by pos-term-to-int-term etc.


(add-program-constant "S" (py "pos=>pos") 1 'const 1)

; (term-to-string (pt "S One"))

(add-program-constant "PosPred" (py "pos=>pos") 1 'const 1)
(add-program-constant "PosHalf" (py "pos=>pos") 1 'const 1)

; We define an overloaded addition.

(add-program-constant "PosPlus" (py "pos=>pos=>pos") 1 'const 2)
(add-program-constant "IntPlus" (py "int=>int=>int") 1 'const 2)
(add-program-constant "RatPlus" (py "rat=>rat=>rat") 1 'const 2)
(add-program-constant "RealPlus" (py "real=>real=>real") 1 'const 2)

(add-token
 "+"
 'add-op
 (lambda (x y)
   (let* ((type1 (term-to-type x))
	  (type2 (term-to-type y))
	  (type (types-lub type1 type2))
	  (type-string
	   (cond ((equal? (py "pos") type) "Pos")
		 ((equal? (py "int") type) "Int")
		 ((equal? (py "rat") type) "Rat")
		 ((equal? (py "real") type) "Real")
		 (else (myerror "+: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Plus")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x y))))

(add-display
 (py "pos")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "PosPlus"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (list 'add-op "+"
	       (term-to-token-tree (car args))
	       (term-to-token-tree (cadr args)))
	 #f))))

(add-display
 (py "int")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "IntPlus"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (list 'add-op "+"
	       (term-to-token-tree (int-term-to-pos-term (car args)))
	       (term-to-token-tree (int-term-to-pos-term (cadr args))))
	 #f))))

(add-display
 (py "rat")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RatPlus"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (list 'add-op "+"
	       (term-to-token-tree
		(int-term-to-pos-term (rat-term-to-int-term (car args))))
	       (term-to-token-tree
		(int-term-to-pos-term (rat-term-to-int-term (cadr args)))))
	 #f))))

(add-display
 (py "real")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if
      (and (term-in-const-form? op)
	   (string=? "RealPlus"
		     (const-to-name (term-in-const-form-to-const op)))
	   (= 2 (length args)))
      (list 'add-op "+"
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term (car args)))))
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term (cadr args))))))
      #f))))

; (term-to-string (pt "(IntN n)+k"))

(add-display
 (py "real")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealConstr"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args))
	      (term-in-abst-form? (car args))
	      (not (member (term-in-abst-form-to-var (car args))
			   (term-to-free
			    (term-in-abst-form-to-kernel (car args))))))
	 (term-to-token-tree
	  (int-term-to-pos-term
	   (rat-term-to-int-term
	    (term-in-abst-form-to-kernel (car args)))))
	 #f))))

; (term-to-string (nt (pt "(IntN n#k)+RealConstr([n]1)([k]7)")))


; We define a cut-off subtraction for positive numbers.

(add-program-constant "PosMinus" (py "pos=>pos=>pos") 1 'const 2)

(add-token
 "--"
 'add-op
 (lambda (x y)
   (mk-term-in-app-form
    (make-term-in-const-form (pconst-name-to-pconst "PosMinus")) x y)))

(add-display
 (py "pos")
 (lambda (x)
   (let* ((op (term-in-app-form-to-final-op x))
	  (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "PosMinus"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (list 'add-op "--"
	       (term-to-token-tree (car args))
	       (term-to-token-tree (cadr args)))
	 #f))))


; We define an overloaded subtraction for integers, rationals and reals.

(add-program-constant "IntMinus" (py "int=>int=>int") 1 'const 2)
(add-program-constant "RatMinus" (py "rat=>rat=>rat") 1 'const 2)
(add-program-constant "RealMinus" (py "real=>real=>real") 1 'const 2)

(add-token
 "-"
 'add-op
 (lambda (x y)
   (let* ((type1 (term-to-type x))
	  (type2 (term-to-type y))
	  (type (types-lub type1 type2))
	  (type-string
	   (cond ((or (equal? (py "pos") type) (equal? (py "int") type)) "Int")
		 ((equal? (py "rat") type) "Rat")
		 ((equal? (py "real") type) "Real")
		 (else (myerror "-: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Minus")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x y))))

(add-display
 (py "int")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "IntMinus"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'add-op "-"
	    (term-to-token-tree (int-term-to-pos-term arg1))
	    (term-to-token-tree (int-term-to-pos-term arg2))))
	 #f))))

(add-display
 (py "rat")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RatMinus"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'add-op "-"
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg1)))
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg2)))))
	 #f))))

(add-display
 (py "real")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealMinus"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'add-op "-"
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg1))))
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg2))))))
	 #f))))


; We define an overloaded multiplication.

(add-program-constant "PosTimes" (py "pos=>pos=>pos") 1 'const 2)
(add-program-constant "IntTimes" (py "int=>int=>int") 1 'const 2)
(add-program-constant "RatTimes" (py "rat=>rat=>rat") 1 'const 2)
(add-program-constant "RealTimes" (py "real=>real=>real") 1 'const 2)

(add-token
 "*"
 'mul-op
 (lambda (x y)
   (let* ((type1 (term-to-type x))
	  (type2 (term-to-type y))
	  (type (types-lub type1 type2))
	  (type-string
	   (cond ((equal? (py "pos") type) "Pos")
		 ((equal? (py "int") type) "Int")
		 ((equal? (py "rat") type) "Rat")
		 ((equal? (py "real") type) "Real")
		 (else (myerror "*: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Times")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x y))))

(add-display
 (py "pos")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "PosTimes"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (list 'mul-op "*"
	       (term-to-token-tree (car args))
	       (term-to-token-tree (cadr args)))
	 #f))))

(add-display
 (py "int")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "IntTimes"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "*"
	    (term-to-token-tree (int-term-to-pos-term arg1))
	    (term-to-token-tree (int-term-to-pos-term arg2))))
	 #f))))

(add-display
 (py "rat")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RatTimes"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "*"
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg1)))
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg2)))))
	 #f))))

(add-display
 (py "real")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealTimes"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "*"
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg1))))
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg2))))))
	 #f))))


; We define an overloaded division for rationals and reals. 

(add-program-constant "RatDiv" (py "rat=>rat=>rat") 1 'const 2)
(add-program-constant "RealDiv" (py "real=>real=>real") 1 'const 2)

(add-token
 "/"
 'mul-op
 (lambda (x y)
   (let* ((type1 (term-to-type x))
	  (type2 (term-to-type y))
	  (type (types-lub type1 type2))
	  (type-string
	   (cond ((equal? (py "pos") type) "Rat")
		 ((equal? (py "int") type) "Rat")
		 ((equal? (py "rat") type) "Rat")
		 ((equal? (py "real") type) "Real")
		 (else (myerror "/: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Div")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x y))))

(add-display
 (py "rat")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RatDiv"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "/"
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg1)))
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg2)))))
	 #f))))

; (term-to-string (pt "n/m"))
; (term-to-string (pt "(IntN k)/m"))

(add-display
 (py "real")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealDiv"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "/"
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg1))))
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg2))))))
	 #f))))

; (term-to-string (pt "x/a"))
; (term-to-string (pt "x/i"))
; (term-to-string (pt "n/a"))
; (type-to-string (term-to-type (pt "n/a")))
; (type-to-string (term-to-type (pt "n/i")))
; (type-to-string (term-to-type (pt "n/x")))


; We define the absolute value for integers, rationals and reals.

(add-program-constant "IntAbs" (py "int=>int") 1 'const 1)
(add-program-constant "RatAbs" (py "rat=>rat") 1 'const 1)
(add-program-constant "RealAbs" (py "real=>real") 1 'const 1)

(add-token
 "abs"
 'prefix-op
 (lambda (x)
   (let* ((type (term-to-type x))
	  (type-string
	   (cond ((equal? (py "int") type) "Int")
		 ((equal? (py "rat") type) "Rat")
		 ((equal? (py "real") type) "Real")
		 (else
		  (myerror "abs: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Abs")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x))))

(add-display
 (py "int")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "IntAbs"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 1 (length args)))
	 (let ((arg (car args)))
	   (list
	    'prefix-op "abs"
	    (term-to-token-tree
	     (int-term-to-pos-term arg))))
	 #f))))

(add-display
 (py "rat")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RatAbs"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 1 (length args)))
	 (let ((arg (car args)))
	   (list
	    'prefix-op "abs"
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg)))))
	 #f))))

; (term-to-string (pt "abs a"))

(add-display
 (py "real")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealAbs"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 1 (length args)))
	 (let ((arg (car args)))
	   (list
	    'prefix-op "abs"
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg))))))
	 #f))))


; The overloaded exponential function, written as operator:

(add-program-constant "PosExp" (py "pos=>pos=>pos") 1 'const 2)
(add-program-constant "IntExp" (py "int=>pos=>int") 1 'const 2)
(add-program-constant "RatExp" (py "rat=>pos=>rat") 1 'const 2)
(add-program-constant "RealExp" (py "real=>pos=>real") 1 'const 2)

(add-token
 "exp"
 'prefix-op
 (lambda (x)
   (let* ((type (term-to-type x))
	  (type-string
	   (cond ((equal? (py "pos") type) "Pos")
		 ((equal? (py "int") type) "Int")
		 ((equal? (py "rat") type) "Rat")
		 ((equal? (py "real") type) "Real")
		 (else (myerror
			"exp: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Exp")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x))))

(add-display
 (py "pos=>pos")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-op x))
	     (arg (term-in-app-form-to-arg x)))
	 (if (and (term-in-const-form? op)
		  (string=? "PosExp"
			    (const-to-name (term-in-const-form-to-const op))))
	     (list 'appterm ""
		   (list 'const "exp")
		   (term-to-token-tree arg))
	     #f))
       #f)))

(add-display
 (py "pos=>int")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-op x))
	     (arg (term-in-app-form-to-arg x)))
	 (if (and (term-in-const-form? op)
		  (string=? "IntExp"
			    (const-to-name (term-in-const-form-to-const op))))
	     (list 'appterm ""
		   (list 'const "exp")
		   (term-to-token-tree
		    (int-term-to-pos-term arg)))
	     #f))
       #f)))

(add-display
 (py "pos=>rat")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-op x))
	     (arg (term-in-app-form-to-arg x)))
	 (if (and (term-in-const-form? op)
		  (string=? "RatExp"
			    (const-to-name (term-in-const-form-to-const op))))
	     (list 'appterm ""
		   (list 'const "exp")
		   (term-to-token-tree
		    (int-term-to-pos-term (rat-term-to-int-term arg))))
	     #f))
       #f)))

; (term-to-string (pt "exp n m"))
; (term-to-string (pt "PosExp n m"))

(add-display
 (py "pos=>real")
 (lambda (x)
   (if (term-in-app-form? x)
       (let ((op (term-in-app-form-to-op x))
	     (arg (term-in-app-form-to-arg x)))
	 (if (and (term-in-const-form? op)
		  (string=? "RealExp"
			    (const-to-name (term-in-const-form-to-const op))))
	     (list 'appterm ""
		   (list 'const "exp")
		   (term-to-token-tree
		    (int-term-to-pos-term
		     (rat-term-to-int-term (real-term-to-rat-term arg)))))
	     #f))
       #f)))


; The overloaded maximum function, written infix:

(add-program-constant "PosMax" (py "pos=>pos=>pos") 1 'const 2)
(add-program-constant "IntMax" (py "int=>int=>int") 1 'const 2)
(add-program-constant "RatMax" (py "rat=>rat=>rat") 1 'const 2)
(add-program-constant "RealMax" (py "real=>real=>real") 1 'const 2)

(add-token
 "max"
 'mul-op
 (lambda (x y)
   (let* ((type1 (term-to-type x))
	  (type2 (term-to-type y))
	  (type (types-lub type1 type2))
	  (type-string
	   (cond ((equal? (py "pos") type) "Pos")
		 ((equal? (py "int") type) "Int")
		 ((equal? (py "rat") type) "Rat")
		 ((equal? (py "real") type) "Real")
		 (else
		  (myerror "max: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Max")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x y))))

(add-display
 (py "pos")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "PosMax"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (list 'mul-op "max"
	       (term-to-token-tree (car args))
	       (term-to-token-tree (cadr args)))
	 #f))))

(add-display
 (py "int")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "IntMax"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "max"
	    (term-to-token-tree (int-term-to-pos-term arg1))
	    (term-to-token-tree (int-term-to-pos-term arg2))))
	 #f))))

(add-display
 (py "rat")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RatMax"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "max"
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg1)))
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg2)))))
	 #f))))

(add-display
 (py "real")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealMax"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "max"
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg1))))
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg2))))))
	 #f))))

; (term-to-string (pt "n max m"))
; (term-to-string (pt "PosMax n m"))


; The overloaded minimum function, written infix:

(add-program-constant "PosMin" (py "pos=>pos=>pos") 1 'const 2)
(add-program-constant "IntMin" (py "int=>int=>int") 1 'const 2)
(add-program-constant "RatMin" (py "rat=>rat=>rat") 1 'const 2)
(add-program-constant "RealMin" (py "real=>real=>real") 1 'const 2)

(add-token
 "min"
 'mul-op
 (lambda (x y)
   (let* ((type1 (term-to-type x))
	  (type2 (term-to-type y))
	  (type (types-lub type1 type2))
	  (type-string
	   (cond ((equal? (py "pos") type) "Pos")
		 ((equal? (py "int") type) "Int")
		 ((equal? (py "rat") type) "Rat")
		 ((equal? (py "real") type) "Real")
		 (else
		  (myerror "min: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Min")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x y))))

(add-display
 (py "pos")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "PosMin"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (list 'mul-op "min"
	       (term-to-token-tree (car args))
	       (term-to-token-tree (cadr args)))
	 #f))))

(add-display
 (py "int")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "IntMin"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "min"
	    (term-to-token-tree (int-term-to-pos-term arg1))
	    (term-to-token-tree (int-term-to-pos-term arg2))))
	 #f))))

(add-display
 (py "rat")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RatMin"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "min"
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg1)))
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg2)))))
	 #f))))

(add-display
 (py "real")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealMin"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'mul-op "min"
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg1))))
	    (term-to-token-tree
	     (int-term-to-pos-term
	      (rat-term-to-int-term (real-term-to-rat-term arg2))))))
	 #f))))

; (term-to-string (pt "n min m"))
; (term-to-string (pt "PosMin n m"))


; We define the intended equivalence relations on rat.  It is
; decidable and hence can be introduced by a program constant.  We
; need an extra token == here, since the standard equality = for
; finitary algebras is available for rat as well.  Equality for reals
; is not decidable.  We view it as a defined predicate constant.  For
; convenience we use the setup of inductively defined predicates,
; although the `inductive' definition is in fact explicit, i.e., does
; not contain recursive calls,

(add-program-constant "RatEq" (py "rat=>rat=>boole") 1 'const 2)

(add-token
 "=="
 'rel-op
 (lambda (x y)
   (let* ((type1 (term-to-type x))
	  (type2 (term-to-type y))
	  (type (types-lub type1 type2))
	  (type-string
	   (if (or (equal? (py "pos") type)
		   (equal? (py "int") type)
		   (equal? (py "rat") type))
	       "Rat"
	       (myerror "==: unexpected type" (type-to-string type))))
	  (internal-name (string-append type-string "Eq")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x y))))

(add-display
 (py "boole")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RatEq"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'rel-op "=="
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg1)))
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg2)))))
	 #f))))

; (term-to-string (pt "n==m"))
; (term-to-string (pt "i==m"))
; (term-to-string (pt "i==a"))


; We define an overloaded less-than relation.  It can be a program
; constant for positive numbers, integers and rationals.  

(add-program-constant "PosLt" (py "pos=>pos=>boole") 1 'const 2)
(add-program-constant "IntLt" (py "int=>int=>boole") 1 'const 2)
(add-program-constant "RatLt" (py "rat=>rat=>boole") 1 'const 2)

(add-token
 "<"
 'rel-op
 (lambda (x y)
   (let* ((type1 (term-to-type x))
	  (type2 (term-to-type y))
	  (type (types-lub type1 type2))
	  (type-string
	   (cond ((equal? (py "pos") type) "Pos")
		 ((equal? (py "int") type) "Int")
		 ((equal? (py "rat") type) "Rat")
		 ((equal? (py "real") type) "Real")
		 (else (myerror "<: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Lt")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x y))))

(add-display
 (py "boole")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (member (const-to-name (term-in-const-form-to-const op))
		      '("PosLt" "IntLt" "RatLt"))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'rel-op "<"
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg1)))
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg2)))))
	 #f))))

; (term-to-string (pt "n<m"))


; We define an overloaded less-than-or-equal relation.

(add-program-constant "PosLe" (py "pos=>pos=>boole") 1 'const 2)
(add-program-constant "IntLe" (py "int=>int=>boole") 1 'const 2)
(add-program-constant "RatLe" (py "rat=>rat=>boole") 1 'const 2)

(add-token
 "<="
 'rel-op
 (lambda (x y)
   (let* ((type1 (term-to-type x))
	  (type2 (term-to-type y))
	  (type (types-lub type1 type2))
	  (type-string
	   (cond ((equal? (py "pos") type) "Pos")
		 ((equal? (py "int") type) "Int")
		 ((equal? (py "rat") type) "Rat")
		 (else (myerror "<=: unexpected type" (type-to-string type)))))
	  (internal-name (string-append type-string "Le")))
     (mk-term-in-app-form
      (make-term-in-const-form (pconst-name-to-pconst internal-name))
      x y))))

(add-display
 (py "boole")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (member (const-to-name (term-in-const-form-to-const op))
		      '("PosLe" "IntLe" "RatLe"))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'rel-op "<="
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg1)))
	    (term-to-token-tree
	     (int-term-to-pos-term (rat-term-to-int-term arg2)))))
	 #f))))

; (term-to-string (pt "n<=m"))


; We introduce an inductively defined predicate RealEq x y by the
; following clause.

(add-ids
 (list (list "RealEq" (make-arity (py "real") (py "real"))))
 '("all x,y.all k abs(x seq(x mod(k+1))-y seq(y mod(k+1)))<=(1#exp 2 k) -> 
            RealEq x y"))

; Notice that we cannot take = and use overloading, because the token
; = has token type rel-op and hence produces a term, not a predicate.

(add-token
 "==="
 'pred-infix
 (lambda (x y)
   (make-predicate-formula (make-idpredconst "RealEq" '() '()) x y)))

(add-idpredconst-display "RealEq" 'pred-infix "===")

(define idpc
  (idpredconst-name-and-types-and-cterms-to-idpredconst "RealEq" '() '()))

; There are no types, since the clauses do not contain type variables,
; and no cterms, since the clauses do not contain parameter predicate
; variables.

(define aconst0 (number-and-idpredconst-to-intro-aconst 0 idpc))
(formula-to-string (aconst-to-formula aconst0))

"all x,y.all k abs(x seq(x mod(k+1))-y seq(y mod(k+1)))<=(1#exp 2 k) -> x===y"

; We now provide some comfort to work with the (general) introduction
; and elimination axioms for the inductively defined predicate RealEq.
; Recall that the (single) introduction axiom can be used in an
; interactive proof with the command (intro 0).

(set-goal
 (pf "all x,y.
       all k abs(x seq(x mod(k+1))-y seq(y mod(k+1)))<=(1#exp 2 k) -> 
       x===y"))
(strip)
(intro 0)
(use 1)
(save "RealEqIntro")

; Recall also that (elim) expects a goal I(ts) -> A[ts].  Then the
; (strengthened) clauses are generated as new goals.

(set-goal
 (pf "all x,y.x===y ->
             all k abs(x seq(x mod(k+1))-y seq(y mod(k+1)))<=(1#exp 2 k)"))
(assume "x" "y")
(elim)
(search)
(save "RealEqElim")

; The following variant will be more useful, because its head will be
; more often of the form of a given goal.

(set-goal
 (pf "all as,M,bs,N.RealConstr as M===RealConstr bs N ->         
                    all k abs(as(M(k+1))-bs(N(k+1)))<=(1#exp 2 k)"))
(strip)
(use-with "RealEqElim"
	  (pt "RealConstr as M") (pt "RealConstr bs N") 1 (pt "k"))
(save "RealEqElimVariant")

; RealPos is a decidable property and hence can be given as a program
; constant.

(add-program-constant "RealPos" (py "real=>pos=>boole") 1 'const 2)

(add-display
 (make-alg "boole")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealPos"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 2 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args)))
	   (list
	    'appterm ""
	    (list
	     'appterm ""
	     (list 'const "RealPos")
	     (term-to-token-tree
	      (int-term-to-pos-term
	       (rat-term-to-int-term (real-term-to-rat-term arg1)))))
	    (term-to-token-tree arg2)))
	 #f))))

(add-computation-rule
 (pt "RealPos(RealConstr as M)k")
 (pt "(1#exp 2 k)<=as(M(k+1))"))

; RealLt is a decidable property and hence can be given as a program
; constant.

(add-program-constant "RealLt" (py "real=>real=>pos=>boole") 1 'const 3)

(add-display
 (make-alg "boole")
 (lambda (x)
   (let ((op (term-in-app-form-to-final-op x))
	 (args (term-in-app-form-to-args x)))
     (if (and (term-in-const-form? op)
	      (string=? "RealLt"
			(const-to-name (term-in-const-form-to-const op)))
	      (= 3 (length args)))
	 (let ((arg1 (car args))
	       (arg2 (cadr args))
	       (arg3 (caddr args)))
	   (list
	    'appterm ""
	    (list
	     'appterm ""
	     (list
	      'appterm ""
	      (list 'const "RealLt")
	      (term-to-token-tree
	       (int-term-to-pos-term
		(rat-term-to-int-term (real-term-to-rat-term arg1)))))
	     (term-to-token-tree
	      (int-term-to-pos-term
	       (rat-term-to-int-term (real-term-to-rat-term arg2)))))
	    (term-to-token-tree arg3)))
	 #f))))

(add-computation-rule
 (pt "RealLt(RealConstr as N)(RealConstr bs M)k")
 (pt "RealPos(RealConstr bs N-RealConstr as M)k"))


; Non-negative reals are defined inductively

(add-ids
 (list (list "RealNNeg" (make-arity (py "real"))))
 '("all x.all k 0<=x seq(x mod k)+(1#exp 2 k) -> RealNNeg x"))

(set-goal (pf "all x. all k 0<=x seq(x mod k)+(1#exp 2 k) -> RealNNeg x"))
(strip)
(intro 0)
(use 1)
(save "RealNNegIntro")

(set-goal (pf "all x.RealNNeg x -> all k 0<=x seq(x mod k)+(1#exp 2 k)"))
(assume "x")
(elim)
(search)
(save "RealNNegElim")

; The following variant will be more useful, because its head will be
; more often of the form of a given goal.

(set-goal
 (pf "all as,M.RealNNeg(RealConstr as M) -> all k 0<=as(M k)+(1#exp 2 k)"))
(strip)
(use-with "RealNNegElim" (pt "RealConstr as M") 1 (pt "k"))
(save "RealNNegElimVariant")

; For reals less-than-or-equal-to is undecidable and hence must be
; treated as an inductively defined predicate.

(add-ids
 (list (list "RealLe" (make-arity (py "real") (py "real"))))
 '("all x,y.RealNNeg(y-x) -> RealLe x y"))

; Notice that we cannot take <= and use overloading, because the token
; <= has token type rel-op and hence produces a term, not a predicate.

(add-token
 "<<="
 'pred-infix
 (lambda (x y)
   (make-predicate-formula (make-idpredconst "RealLe" '() '()) x y)))

(add-idpredconst-display "RealLe" 'pred-infix "<<=")

(define idpc
  (idpredconst-name-and-types-and-cterms-to-idpredconst "RealLe" '() '()))

; There are no types, since the clauses do not contain type variables,
; and no cterms, since the clauses do not contain parameter predicate
; variables.

(define aconst0 (number-and-idpredconst-to-intro-aconst 0 idpc))
(formula-to-string (aconst-to-formula aconst0))
; "all x,y.RealNNeg(y-x) -> x<<=y"

; We now provide some comfort to work with the (general) introduction
; and elimination axioms for the inductively defined predicate RealLe.
; Recall that the (single) introduction axiom can be used in an
; interactive proof with the command (intro 0).

(set-goal
 (pf "all x,y.RealNNeg(y-x) -> x<<=y"))
(strip)
(intro 0)
(use 1)
(save "RealLeIntro")

; Recall also that (elim) expects a goal I(ts) -> A[ts].  Then the
; (strengthened) clauses are generated as new goals.

(set-goal
 (pf "all x,y.x<<=y -> RealNNeg(y-x)"))
(assume "x" "y")
(elim)
(search)
(save "RealLeElim")

; The following variant will be useful as well, when its head is of
; the form of a given goal.  However, the proof below assumes
;  (add-computation-rule
;   (pt "RealConstr as M-RealConstr bs N")
;   (pt "RealConstr([n]as n-bs n)([k]M(k+1)max N(k+1))"))
; which will only be added later.

; (set-goal
;  (pf "all as,M,bs,N.RealConstr as M<<=RealConstr bs N ->         
;                    RealNNeg(RealConstr([n]bs n-as n)([k]N(k+1)max M(k+1)))"))
; (strip)
; (use-with "RealLeElim" (pt "RealConstr as M") (pt "RealConstr bs N") 1)
; (save "RealLeElimVariant")


; 3. Arithmetic
; =============

(add-computation-rule (pt "S One") (pt "SZero One"))
(add-computation-rule (pt "S(SZero n)") (pt "SOne n"))
(add-computation-rule (pt "S(SOne n)") (pt "SZero(S n)"))

(add-computation-rule (pt "PosPred One") (pt "One"))
(add-computation-rule (pt "PosPred(SZero One)") (pt "One"))
(add-computation-rule
 (pt "PosPred(SZero(SZero n))") (pt "SOne(PosPred(SZero n))"))
(add-computation-rule (pt "PosPred(SZero(SOne n))") (pt "SOne(SZero n)"))
(add-computation-rule (pt "PosPred(SOne n)") (pt "SZero n"))

; (display-program-constants "PosPred")

(add-computation-rule (pt "PosHalf One") (pt "1"))
(add-computation-rule (pt "PosHalf(SZero n)") (pt "n"))
(add-computation-rule (pt "PosHalf(SOne n)") (pt "n"))

; (display-program-constants "PosHalf")

(add-computation-rule (pt "n+One") (pt "S n"))
(add-computation-rule (pt "One+SZero n") (pt "SOne n"))
(add-computation-rule (pt "SZero n+SZero m") (pt "SZero(n+m)"))
(add-computation-rule (pt "SOne n+SZero m") (pt "SOne(n+m)"))
(add-computation-rule (pt "One+SOne n") (pt "SZero(S n)"))
(add-computation-rule (pt "SZero n+SOne m") (pt "SOne(n+m)"))
(add-computation-rule (pt "SOne n+SOne m") (pt "SZero(S(n+m))"))

; (term-to-string (nt (pt "4019+188")))

; (set-goal (pf "all n One+n=S n"))
; (cases)
; (auto)
(add-rewrite-rule (pt "One+n") (pt "S n"))

; (set-goal (pf "all n,m S n+m=S(n+m)"))
; (ind)
; (ind)
; (auto)

; ; ?_3: all n57.all m S n57+m=S(n57+m) -> all m S(SZero n57)+m=S(SZero n57+m)
; (assume "n" "IHzero")
; (ind)
; (auto)

; ; ?_4: all n58.all m S n58+m=S(n58+m) -> all m S(SOne n58)+m=S(SOne n58+m)
; (assume "n" "IHone")
; (ind)
; (auto)

(add-rewrite-rule (pt "n+SZero m") (pt "n+m+m"))
(add-rewrite-rule (pt "n+SOne m") (pt "n+m+m+1"))
; (add-rewrite-rule (pt "SZero n+m") (pt "n+n+m")) ;does not terminate for 2+m
; (add-rewrite-rule (pt "SOne n+m") (pt "n+n+1+m"))

(add-rewrite-rule (pt "n1+(n2+n3)") (pt "n1+n2+n3"))

(add-rewrite-rule (pt "S n+m") (pt "S(n+m)"))
(add-rewrite-rule (pt "n+S m") (pt "S(n+m)"))

; (display-program-constants "PosPlus")


; Computation rules for PosMinus:  They give correct results (only) if m<n.

(add-computation-rule (pt "1--n") (pt "1"))
(add-computation-rule (pt "SZero n--1") (pt "PosPred(SZero n)"))
(add-computation-rule (pt "SZero n--SZero m") (pt "SZero(n--m)"))
(add-computation-rule (pt "SZero n--SOne m") (pt "PosPred(SZero(n--m))"))
(add-computation-rule (pt "SOne n--1") (pt "SZero n"))
(add-computation-rule (pt "SOne n--SZero m") (pt "SZero n--(SZero m--1)"))
;(add-computation-rule (pt "SOne n--SZero m") (pt "[if (m<n) (SOne(n--m)) 1]"))
(add-computation-rule (pt "SOne n--SOne m") (pt "SZero(n--m)"))

; (display-program-constants "PosMinus")

; (term-to-string (nt (pt "6--4")))
; (term-to-string (nt (pt "66--44")))

; Computation rules for PosTimes:

(add-computation-rule (pt "n*One") (pt "n"))
(add-computation-rule (pt "n*SZero m") (pt "SZero(n*m)"))
(add-computation-rule (pt "n*SOne m") (pt "SZero(n*m)+n"))

; (set-goal (pf "all n One*n=n"))
; (ind)
; (auto)
(add-rewrite-rule (pt "One*n") (pt "n"))

; (set-goal (pf "all n,m SZero n*m=SZero(n*m)"))
; (assume "n")
; (ind)
; (auto)

; (assume "m" "IHone")
; (ng)
; (simp "IHone")
; (use "Truth-Axiom")
(add-rewrite-rule (pt "SZero n*m") (pt "SZero(n*m)"))

(add-rewrite-rule (pt "SOne n*m") (pt "SZero(n*m)+m"))
(add-rewrite-rule (pt "n1*(n2*n3)") (pt "n1*n2*n3"))

; (display-program-constants "PosTimes")


; Computation rules for PosExp

(add-computation-rule (pt "exp n One") (pt "n"))

; The computation rule exp n(SZero m) -> (exp n m)*(exp n m) leads to
; unwanted unfoldings.


; Computation rules for PosMax

(add-computation-rule (pt "One max n") (pt "n"))
(add-computation-rule (pt "SZero n max One") (pt "SZero n"))
(add-computation-rule (pt "SZero n max SZero m") (pt "SZero(n max m)"))
(add-computation-rule (pt "SZero n max SOne m")
		      (pt "[if (n<=m) (SOne m) (SZero n)]"))
(add-computation-rule (pt "SOne n max One") (pt "SOne n"))
(add-computation-rule (pt "SOne n max SZero m")
		      (pt "[if (m<=n) (SOne n) (SZero m)]"))
(add-computation-rule (pt "SOne n max SOne m") (pt "SOne(n max m)"))

(add-rewrite-rule (pt "n max 1") (pt "n"))
(add-rewrite-rule (pt "n max n") (pt "n"))

; (term-to-string (nt (pt "7 max 3")))


; Computation rules for PosMin

(add-computation-rule (pt "One min n") (pt "One"))
(add-computation-rule (pt "SZero n min One") (pt "One"))
(add-computation-rule (pt "SZero n min SZero m") (pt "SZero(n min m)"))
(add-computation-rule (pt "SZero n min SOne m")
		      (pt "[if (n<=m) (SZero n) (SOne m)]"))
(add-computation-rule (pt "SOne n min One") (pt "One"))
(add-computation-rule (pt "SOne n min SZero m")
		      (pt "[if (n<=m) (SZero m) (SOne n)]"))
(add-computation-rule (pt "SOne n min SOne m") (pt "SOne(n min m)"))

; (term-to-string (nt (pt "7 min 3")))


; Computation rules for PosLt, PosLe

(add-computation-rule (pt "n<One") (pt "False"))
(add-computation-rule (pt "One<SZero n") (pt "True"))
(add-computation-rule (pt "One<SOne n") (pt "True"))
(add-computation-rule (pt "SZero n<SZero m") (pt "n<m"))
(add-computation-rule (pt "SZero n<SOne m") (pt "n<=m"))
(add-computation-rule (pt "SOne n<SZero m") (pt "n<m"))
(add-computation-rule (pt "SOne n<SOne m") (pt "n<m"))

(add-computation-rule (pt "One<=n") (pt "True"))
(add-computation-rule (pt "SZero n<=One") (pt "False"))
(add-computation-rule (pt "SOne n<=One") (pt "False"))
(add-computation-rule (pt "SZero n<=SZero m") (pt "n<=m"))
(add-computation-rule (pt "SZero n<=SOne m") (pt "n<=m"))
(add-computation-rule (pt "SOne n<=SZero m") (pt "n<m"))
(add-computation-rule (pt "SOne n<=SOne m") (pt "n<=m"))

; (term-to-string (nt (pt "27<27")))
; (term-to-string (nt (pt "27<=27")))

(add-rewrite-rule (pt "n<n") (pt "False"))
(add-rewrite-rule (pt "n<S n") (pt "True"))
(add-rewrite-rule (pt "S n<n") (pt "False"))
(add-rewrite-rule (pt "n<n+m") (pt "True"))
(add-rewrite-rule (pt "S n<S m") (pt "n<m"))
(add-rewrite-rule (pt "1<S n") (pt "True"))

(add-rewrite-rule (pt "n<=n") (pt "True"))
(add-rewrite-rule (pt "n<=S n") (pt "True"))
(add-rewrite-rule (pt "S n<=n") (pt "False"))
(add-rewrite-rule (pt "n<=n+m") (pt "True"))
(add-rewrite-rule (pt "S n<=S m") (pt "n<=m"))
(add-rewrite-rule (pt "S n<=1") (pt "False"))


; Computation rules for IntPlus

(add-computation-rule (pt "0+i") (pt "i"))
(add-computation-rule (pt "IntP n+0") (pt "IntP n"))
(add-computation-rule (pt "IntP n+IntP m") (pt "IntP(n+m)"))
(add-computation-rule (pt "IntP n+IntN m")
		      (pt "[if (n=m) 0 [if (n<m) (IntN(m--n)) (IntP(n--m))]]"))
(add-computation-rule (pt "IntN n+0") (pt "IntN n"))
(add-computation-rule (pt "IntN n+IntP m")
		      (pt "[if (n=m) 0 [if (n<m) (IntP(m--n)) (IntN(n--m))]]"))
(add-computation-rule (pt "IntN n+IntN m") (pt "IntN(n+m)"))

(add-rewrite-rule (pt "i+0") (pt "i"))
		      
; Computation rules for IntMinus

(add-computation-rule (pt "i-0") (pt "i"))
(add-computation-rule (pt "IntP n-IntP m")
		      (pt "[if (n=m) 0 [if (n<m) (IntN(m--n)) (IntP(n--m))]]"))
(add-computation-rule (pt "IntP n-IntN m") (pt "IntP(n+m)"))
(add-computation-rule (pt "IntN n-IntP m") (pt "IntN(n+m)"))
(add-computation-rule (pt "IntN n-IntN m")
		      (pt "[if (n=m) 0 [if (n<m) (IntP(m--n)) (IntN(n--m))]]"))
(add-computation-rule (pt "0-IntN n") (pt "IntP n"))
(add-computation-rule (pt "0-IntP n") (pt "IntN n"))
		      
; Computation rules for IntTimes

(add-computation-rule (pt "0*i") (pt "0"))
(add-computation-rule (pt "IntP n*0") (pt "0"))
(add-computation-rule (pt "IntP n*IntP m") (pt "IntP(n*m)"))
(add-computation-rule (pt "IntP n*IntN m") (pt "IntN(n*m)"))
(add-computation-rule (pt "IntN n*0") (pt "0"))
(add-computation-rule (pt "IntN n*IntP m") (pt "IntN(n*m)"))
(add-computation-rule (pt "IntN n*IntN m") (pt "IntP(n*m)"))

(add-rewrite-rule (pt "i*0") (pt "0"))
(add-rewrite-rule (pt "i*1") (pt "i"))
(add-rewrite-rule (pt "1*i") (pt "i"))

; Computation rules for IntAbs

(add-computation-rule (pt "abs 0") (pt "0"))
(add-computation-rule (pt "abs IntP n") (pt "IntP n"))
(add-computation-rule (pt "abs IntN n") (pt "IntP n"))

; Computation rules for IntExp

(add-computation-rule (pt "exp i One") (pt "i"))

; Computation rules for IntMax

(add-computation-rule (pt "0 max 0") (pt "0"))
(add-computation-rule (pt "0 max IntP m") (pt "IntP m"))
(add-computation-rule (pt "0 max IntN m") (pt "0"))
(add-computation-rule (pt "IntP n max 0") (pt "IntP n"))
(add-computation-rule (pt "IntP n max IntP m") (pt "IntP(n max m)"))
(add-computation-rule (pt "IntP n max IntN m") (pt "IntP n"))
(add-computation-rule (pt "IntN n max 0") (pt "0"))
(add-computation-rule (pt "IntN n max IntP m") (pt "IntP m"))
(add-computation-rule (pt "IntN n max IntN m") (pt "IntN(n min m)"))

; Computation rules for IntMin

(add-computation-rule (pt "0 min 0") (pt "0"))
(add-computation-rule (pt "0 min IntP m") (pt "0"))
(add-computation-rule (pt "0 min IntN m") (pt "IntN m"))
(add-computation-rule (pt "IntP n min 0") (pt "0"))
(add-computation-rule (pt "IntP n min IntP m") (pt "IntP(n min m)"))
(add-computation-rule (pt "IntP n min IntN m") (pt "IntN m"))
(add-computation-rule (pt "IntN n min 0") (pt "IntN n"))
(add-computation-rule (pt "IntN n min IntP m") (pt "IntN n"))
(add-computation-rule (pt "IntN n min IntN m") (pt "IntN(n max m)"))

; Computation rules for IntLt

(add-computation-rule (pt "0<0") (pt "False"))
(add-computation-rule (pt "0<IntP m") (pt "True"))
(add-computation-rule (pt "0<IntN m") (pt "False"))
(add-computation-rule (pt "IntP n<0") (pt "False"))
(add-computation-rule (pt "IntP n<IntP m") (pt "n<m"))
(add-computation-rule (pt "IntP n<IntN m") (pt "False"))
(add-computation-rule (pt "IntN n<0") (pt "True"))
(add-computation-rule (pt "IntN n<IntP m") (pt "True"))
(add-computation-rule (pt "IntN n<IntN m") (pt "m<n"))

; (term-to-string (nt (pt "2<IntN 8")))

; Computation rules for IntLe

(add-computation-rule (pt "0<=0") (pt "True"))
(add-computation-rule (pt "0<=IntP m") (pt "True"))
(add-computation-rule (pt "0<=IntN m") (pt "False"))
(add-computation-rule (pt "IntP n<=0") (pt "False"))
(add-computation-rule (pt "IntP n<=IntP m") (pt "n<=m"))
(add-computation-rule (pt "IntP n<=IntN m") (pt "False"))
(add-computation-rule (pt "IntN n<=0") (pt "True"))
(add-computation-rule (pt "IntN n<=IntP m") (pt "True"))
(add-computation-rule (pt "IntN n<=IntN m") (pt "m<=n"))

; (term-to-string (nt (pt "2<=IntP 2")))

(add-rewrite-rule (pt "i<i") (pt "False"))
(add-rewrite-rule (pt "i<i+m") (pt "True"))

(add-rewrite-rule (pt "i<=i") (pt "True"))
(add-rewrite-rule (pt "i<=i+m") (pt "True"))


; Computation rules for RatPlus

(add-computation-rule (pt "(i1#k1)+(i2#k2)") (pt "i1*k2+i2*k1#k1*k2"))

; Computation rules for RatMinus

(add-computation-rule (pt "(i1#k1)-(i2#k2)") (pt "i1*k2-i2*k1#k1*k2"))

; Computation rules for RatTimes

(add-computation-rule (pt "(i1#k1)*(i2#k2)") (pt "i1*i2#k1*k2"))

; Computation rules for RatDiv

(add-computation-rule (pt "(i1#k1)/(IntP n2#k2)") (pt "(i1*k2#n2*k1)"))
(add-computation-rule (pt "(i1#k1)/(IntN n2#k2)") (pt "((0-i1)*k2#n2*k1)"))

; Computation rules for RatAbs

(add-computation-rule (pt "abs(0#k)") (pt "0#k"))
(add-computation-rule (pt "abs(IntP n#k)") (pt "IntP n#k"))
(add-computation-rule (pt "abs(IntN n#k)") (pt "IntP n#k"))

; Computation rules for RatExp

(add-computation-rule (pt "exp a One") (pt "a"))

; Computation rules for RatMax

(add-computation-rule (pt "(i1#k1)max(i2#k2)")
		      (pt "[if (i1*k2<=i2*k1) (i2#k2) (i1#k1)]"))

; Computation rules for RatMin

(add-computation-rule (pt "(i1#k1)min(i2#k2)")
		      (pt "[if (i1*k2<=i2*k1) (i1#k1) (i2#k2)]"))

; Computation rules for RatLe

(add-computation-rule (pt "(i1#k1)<=(i2#k2)") (pt "i1*k2<=i2*k1"))

; Computation rules for RatLt

(add-computation-rule (pt "(i1#k1)<(i2#k2)") (pt "i1*k2<i2*k1"))

(add-rewrite-rule (pt "a<a") (pt "False"))
(add-rewrite-rule (pt "a<a+m") (pt "True"))

(add-rewrite-rule (pt "a<=a") (pt "True"))
(add-rewrite-rule (pt "a<=a+m") (pt "True"))
(add-rewrite-rule (pt "0<=(IntN n#m)") (pt "False"))

; Computation rules for RatEq

(add-computation-rule (pt "(i1#k1)==(i2#k2)") (pt "i1*k2=i2*k1"))

(add-rewrite-rule (pt "a==a") (pt "True"))


; Computation rules for RealPlus

(add-computation-rule
 (pt "RealConstr as M+RealConstr bs N")
 (pt "RealConstr([n]as n+bs n)([k]M(k+1)max N(k+1))"))

; Computation rules for RealMinus

(add-computation-rule
 (pt "RealConstr as M-RealConstr bs N")
 (pt "RealConstr([n]as n-bs n)([k]M(k+1)max N(k+1))"))

; Computation rules for RealAbs

(add-computation-rule
 (pt "abs(RealConstr as M)") (pt "RealConstr([n]abs(as n))M"))
